#include "ELF.h"
#include "InitRD.h"
#include "Log.h"
#include "arch/CPU.h"
#include "arch/MMU.h"
#include "arch/Timer.h"
#include "boot/Init.h"
#include "config.h"
#include "memory/Heap.h"
#include "memory/KernelVM.h"
#include "memory/MemoryManager.h"
#include "thread/Scheduler.h"
#include <luna/CString.h>
#include <luna/Result.h>
#include <luna/Units.h>

void heap_thread()
{
    CPU::disable_interrupts();
    dump_heap_usage();
    kdbgln("Kernel uses %lu vm pages", KernelVM::used() / ARCH_PAGE_SIZE);
    kernel_exit();
}

void reap_thread()
{
    while (true)
    {
        CPU::disable_interrupts();
        auto dying_threads = Scheduler::check_for_dying_threads();
        CPU::enable_interrupts();

        dying_threads.consume([](Thread* thread) { Scheduler::reap_thread(thread); });

        kernel_sleep(250);
    }
}

Result<void> init()
{
    kinfoln("Starting Moon %s, built on %s at %s", MOON_VERSION, __DATE__, __TIME__);

    kinfoln("Current platform: %s", CPU::platform_string());

    kinfoln("Current processor: %s", CPU::identify().value_or("(unknown)"));

    Timer::init();

    kinfoln("Total memory:    %s", to_dynamic_unit(MemoryManager::total()).release_value().chars());
    kinfoln("Free memory:     %s", to_dynamic_unit(MemoryManager::free()).release_value().chars());
    kinfoln("Used memory:     %s", to_dynamic_unit(MemoryManager::used()).release_value().chars());
    kinfoln("Reserved memory: %s", to_dynamic_unit(MemoryManager::reserved()).release_value().chars());

    Thread::init();
    Scheduler::init();

    TarStream::Entry entry;
    while (TRY(g_initrd.read_next_entry().try_set_value_with_specific_error(entry, 0)))
    {
        if (entry.type == TarStream::EntryType::RegularFile)
        {
            kinfoln("Found file %s in initial ramdisk, of size %s", entry.name,
                    to_dynamic_unit(entry.size).release_value().chars());

            if (!strcmp(entry.name, "bin/app")) { TRY(Scheduler::new_userspace_thread(entry, g_initrd)); }
        }
    }

    TRY(Scheduler::new_kernel_thread(heap_thread));
    TRY(Scheduler::new_kernel_thread(reap_thread));

    CPU::platform_finish_init();

    CPU::enable_interrupts();

    return {};
}

extern "C" [[noreturn]] void _start()
{
    Init::check_magic();
    Init::early_init();
    auto rc = init();
    if (rc.has_error()) kerrorln("Runtime error: %s", rc.error_string());
    CPU::idle_loop();
}
